# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Global fig config utils."""

from __future__ import annotations

import dataclasses
import typing
from typing import Callable, Generic, Optional, TypeVar

from etils import edc

_T = TypeVar('_T')


@edc.dataclass
@dataclasses.dataclass
class FigConfig:
  """Figure configuration options.

  Can be mutated globally, like:

  ```python
  v3d.fig_config.show_zero = False
  v3d.fig_config.num_samples_point3d = None  # Do not subsample point display
  ```

  Attributes:
    show_zero: Whether to show the `(0, 0, 0)` origin, otherwise the plot x, y,
      z axis adapt to the data.
    num_samples_point3d: Max number of `v3d.Point3d` displayed by default (
      `None` for all)
    num_samples_point2d: Max number of `v3d.Point2d` displayed by default (
      `None` for all)
    num_samples_ray: Max number of `v3d.Ray` displayed by default (`None` for
      all)
    cam_scale: Scale of the cameras.
  """

  # When updating this, also update `v3d.make_fig` (for auto-complete /
  # discoverability)

  show_zero: bool = True
  num_samples_point3d: Optional[int] = 10_000
  num_samples_point2d: Optional[int] = 50_000
  num_samples_ray: Optional[int] = 500
  cam_scale: float = 1.0

  def replace(self: _T, **kwargs) -> _T:
    # Filter `...` (forwarded from `v3d.make_fig`)
    kwargs = {k: v for k, v in kwargs.items() if v is not ...}
    return dataclasses.replace(self, **kwargs)


fig_config = FigConfig()


@edc.dataclass(kw_only=True)
@dataclasses.dataclass(frozen=True)
class TraceConfig:
  """Configuration of a single v3d object.

  Attributes:
    name: The name of the figure.
    num_samples: Maximum number of X to display (`-1` to display all). Keep
      rendering time reasonable by displaying only a subset of the total X.
  """

  # NOTE: When adding new properties here, please also update all
  # `.replace_fig_config(` function to get type checking/auto-complete.

  name: Optional[str] = None
  num_samples: Optional[int] = None

  # Hidden reference to the global `fig_config`.
  # This allow to locally overwrite the default `fig_config`, like:
  # `v3d.make_fig(rays, num_samples_ray=None)`
  # This is used for the lazy values
  _fig_config: FigConfig = dataclasses.field(default=fig_config, repr=False)

  if not typing.TYPE_CHECKING:

    def __getattribute__(self, name: str):
      # Auto-resolve lazy values
      value = super().__getattribute__(name)
      if isinstance(value, LazyValue):
        return value.get_value(self._fig_config)
      else:
        return value

  def replace(self: _T, **kwargs) -> _T:
    """Alias for `dataclasses.replace`."""
    # Use a custom replace because `dataclasses.replace` remove the `LazyValue`

    def _get_attr(name: str) -> _T:
      if name in kwargs:
        return kwargs[name]
      else:
        return super(TraceConfig, self).__getattribute__(name)

    init_kwargs = {
        f.name: _get_attr(f.name) for f in dataclasses.fields(self) if f.init
    }
    return type(self)(**init_kwargs)


class LazyValue(Generic[_T]):
  """Lazyly compute a value."""

  def __init__(self, value_fn: Callable[[FigConfig], _T]):
    self._value_fn = value_fn

  def get_value(self, fig_config_: FigConfig) -> _T:
    return self._value_fn(fig_config_)

  def __repr__(self) -> str:
    # Technically, this should be the `object._fig_config` on which the
    # `LazyValue` is attached. Could be implemented with descriptor.
    return f'{type(self).__name__}({self._value_fn(fig_config)})'



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotly utils."""

from __future__ import annotations

import abc
from collections.abc import Sequence  # pylint: disable=g-importing-member
from typing import Any, Dict, List, Optional, Union

import dataclass_array as dca
import einops
from etils import enp
from etils.array_types import Array, FloatArray  # pylint: disable=g-multiple-import
import numpy as np
from visu3d import array_dataclass as v3d_dataclass_array
from visu3d import math
from visu3d.plotly import fig_config_utils
from visu3d.plotly import traces_builder
from visu3d.utils import py_utils
from visu3d.utils.lazy_imports import plotly_base
from visu3d.utils.lazy_imports import plotly_go as go

_Primitive = Union[str, int, bool, float]
_PlotlyKwargs = Dict[str, Union[np.ndarray, _Primitive]]

del abc  # TODO(epot): Why pytype doesn't like abc.ABC ?

_MARKER_SYMBOLS = {
    'circle',
    'circle-open',
    'cross',
    'diamond',
    'diamond-open',
    'square',
    'square-open',
    'x',
}

# TODO(epot): More dynamic sub-sampling for points ?
# * controled in `v3d.make_fig`
# * globally assigned (collect the global batch shape)
# * Add a tqdm bar ?

# TODO(epot): Refactor this file in 3 smaller files:
# * `traces_utils.py`: The `make_xyz` functions
# * `fig_utils.py`: The `make_fig` function
# * `visualization`: `Visualizable` & typing annotations

# Any class with the `make_traces` protocol
# TODO(epot): Could use protocol instead
VisualizableInterface = Any


class Visualizable:  # (abc.ABC):
  """Interface for elements which are visualizable."""

  # @abc.abstractmethod
  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    """Construct the traces of the given object."""
    raise NotImplementedError(
        f'To be visualizable, {type(self).__qualname__} should implement '
        'the `.make_traces()` protocol.'
    )

  @property
  def fig(self) -> go.Figure:
    """Construct the figure of the given object."""
    return make_fig([self])


VisualizableItem = Union[Visualizable, Array[...], VisualizableInterface]
VisualizableArg = Union[VisualizableItem, List[VisualizableItem]]


# TODO(epot): Potential improvement:
# * Accept a `dict[str, data]` (to have trace name)
# * Allow wrapping data in some `v3d.FigData(data, **kwargs)` to allow
#   customize metadata (color, point name,...) ?
# * Allow nested structure to auto-group multiple traces ?
def make_fig(
    *data: VisualizableArg,
    # Those arguments match `v3d.fig_config` options
    # They are duplicated for discoverability / auto-complete
    show_zero: bool = ...,  # pytype: disable=annotation-type-mismatch
    num_samples_point3d: Optional[int] = ...,  # pytype: disable=annotation-type-mismatch
    num_samples_point2d: Optional[int] = ...,  # pytype: disable=annotation-type-mismatch
    num_samples_ray: Optional[int] = ...,  # pytype: disable=annotation-type-mismatch
    cam_scale: float = ...,  # pytype: disable=annotation-type-mismatch
    **fig_config_kwargs: Any,
) -> go.Figure:
  """Returns the figure from the given data.

  ```python
  v3d.make_fig([obj0, obj1])  # Or `v3d.make_fig(obj0, obj1)`
  ```

  The figure will use `v3d.fig_config` global figure configuration. Individual
  config options can be overwritten here through kwargs.

  Args:
    *data: The data to plot. Either a `v3d.Vizualizable` or a `np.array` point
      cloud, or a list of the above.
    show_zero: Whether to show the `(0, 0, 0)` origin, otherwise the plot x, y,
      z axis adapt to the data.
    num_samples_point3d: Max number of `v3d.Point3d` displayed by default (
      `None` for all)
    num_samples_point2d: Max number of `v3d.Point2d` displayed by default (
      `None` for all)
    num_samples_ray: Max number of `v3d.Ray` displayed by default (`None` for
      all)
    cam_scale: Scale of the cameras.
    **fig_config_kwargs: Additional figure options (see `v3d.fig_config`)

  Returns:
    The plotly `go.Figure`, can be further modified.
  """
  traces = make_traces(
      *data,
      show_zero=show_zero,
      num_samples_point3d=num_samples_point3d,
      num_samples_point2d=num_samples_point2d,
      num_samples_ray=num_samples_ray,
      cam_scale=cam_scale,
      **fig_config_kwargs,
  )
  fig = go.Figure(data=traces)

  is_2d = _is_traces_2d(traces)

  if not _show_legend(traces):  # Eventually disable the legend
    fig.update_traces(showlegend=False)

  fig.update_layout(
      margin=dict(l=0, r=0, t=0, b=0),
      legend=dict(
          y=0.96,  # Move legend bellow to avoid overlapp with title
      ),
  )

  if is_2d:
    fig.update_yaxes(
        scaleanchor='x',
        scaleratio=1,
        autorange='reversed',
    )
  else:
    # TODO(epot): To avoid axes to flicker when toogle a trace, we should
    # fix the range, ratio,... of the axes to the max data boundaries.
    fig.update_scenes(
        aspectmode='data',  # Keep equal axis
        # TODO(plotly): Ideally, should set `xaxis_rangemode` rather than
        # having to add `make_zero_point`. But plotly is buggy.
        # Extend axis to include 0
        # xaxis_rangemode='tozero',
        # yaxis_rangemode='tozero',
        # zaxis_rangemode='tozero',
    )
  return fig


def make_traces(
    *data: VisualizableArg,
    **fig_config_kwargs: Any,
) -> list[plotly_base.BaseTraceType]:
  """Returns the traces from the given data."""
  #
  if len(data) == 1:  # `v3d.make_traces([a, b, c])` or `v3d.make_traces(a)`
    (data,) = data
    if not isinstance(data, (tuple, list)):
      data = [data]
  # Otherwise, called as `v3d.make_traces(a, b, c)`

  # TODO(epot): `curr_config` should be created in `make_fig` and
  # propagated downstream.
  curr_config = fig_config_utils.fig_config.replace(**fig_config_kwargs)

  # Mapping to count the number of trace of a type
  trace_namer = traces_builder.TraceNamer()

  # TODO(epot): Should dynamically sub-sample across all traces, instead of
  # subsampling individual traces.
  traces = []
  for val in data:
    if is_visualizable(val):
      if isinstance(val, dca.DataclassArray):
        if has_fig_config(val):
          # Overwrite the global fig_config with the local copy
          if isinstance(val, v3d_dataclass_array.DataclassArray):
            val = val.replace_fig_config(_fig_config=curr_config)
          val = math.subsample(
              val,
              num_samples=val.fig_config.num_samples,
              # TODO(epot): Should likely make the seed depend on other
              # factors like class name, position in make_traces *args,...
              seed=0,
          )
        val = val.as_np()
      sub_traces = val.make_traces()  # pytype: disable=attribute-error
      # Normalizing trace
      if isinstance(sub_traces, plotly_base.BaseTraceType):
        sub_traces = [sub_traces]
      trace_namer.set_name(sub_traces, val)
      traces.extend(sub_traces)
    elif enp.lazy.is_array(val) or isinstance(val, list):
      val = np.asarray(val)
      sub_traces = make_points(val, num_samples=curr_config.num_samples_point3d)
      trace_namer.set_name(sub_traces, val)
      traces.extend(sub_traces)
    elif isinstance(val, plotly_base.BaseTraceType):  # Already a trace
      traces.append(val)
    else:
      raise TypeError(f'Unsuported {type(val)}')
  if curr_config.show_zero and not _is_traces_2d(traces):
    traces.append(make_zero_point())
  return traces


def make_points(
    coords: FloatArray['*d num_dim'],
    *,
    color: Array['*d channel'] = None,
    num_samples: Optional[int] = None,
) -> list[plotly_base.BaseTraceType]:
  """Display a 2d or 3d point cloud.

  Args:
    coords: The point coordinates
    color: The optional point colors. Can be RGB uint8 value, or scalar float
      scale.
    num_samples: Returns the number of samples

  Returns:
    The plotly trace to display
  """
  if coords.shape[-1] not in (2, 3):
    raise ValueError(
        'Points should be `(..., 2)` or `(..., 3)`. '
        f'now. Got shape={coords.shape}'
    )
  if color is not None:
    assert color.shape[:-1] == coords.shape[:-1]
    color = color.reshape((-1, color.shape[-1]))
  coords = coords.reshape((-1, coords.shape[-1]))

  # TODO(epot): Subsample array if nb points >500
  coords, color = subsample(coords, color, num_samples=num_samples)  # pylint: disable=unbalanced-tuple-unpacking

  if color is not None:
    color = _normalize_color(color)

  if coords.shape[-1] == 3:
    point_cloud = _make_scatter_3d(coords, color=color)
  elif coords.shape[-1] == 2:
    point_cloud = _make_scatter_2d(coords, color=color)
  else:
    raise AssertionError
  return [point_cloud]


def _normalize_color(color: Array['*d channel']):
  """Normalize the color to plotly compatible arg."""
  if color.shape[-1] == 3:
    return [f'rgb({r}, {g}, {b})' for r, g, b in color]
  elif color.shape[-1] == 1:
    return color
  else:
    raise ValueError('Color should be (..., 1) or (..., 3)')


def _make_scatter_3d(
    coords: FloatArray['*d 3'],
    *,
    color: Array['*d channel'] = None,
) -> go.Scatter3d:
  """Returns 3d scatter plots."""
  points_xyz_kwargs = to_xyz_dict(coords)
  return go.Scatter3d(
      **points_xyz_kwargs,
      mode='markers',
      marker=go.scatter3d.Marker(
          size=2.0,
          color=color,
      ),
  )


def _make_scatter_2d(
    coords: FloatArray['*d 2'],
    *,
    color: Array['*d channel'] = None,
) -> go.Scattergl:
  """Returns 2d scatter plots."""
  # TODO(epot): Correct axis (x, y) to match numpy !!
  # Currently plotly use the `(y, x)` convention. I could not find a way to
  # invert
  points_xy_kwargs = to_xyz_dict(coords, names='xy')
  return go.Scattergl(
      **points_xy_kwargs,
      mode='markers',
      marker=go.scattergl.Marker(
          size=2.0,
          color=color,
      ),
  )


def make_lines_traces(
    start: FloatArray['*lines 3'],
    end: FloatArray['*lines 3'],
    *,
    axis: int = -1,
    end_marker: Optional[str] = None,
) -> list[plotly_base.BaseTraceType]:
  """Trace independent lines.

  Args:
    start: (x, y, z) coordinates of the start points
    end: (x, y, z) coordinates of the end points
    axis: Axis on which the (x, y, z) coordinates are defined
    end_marker: Marker at the end line. Can be any `go.Scatter3d.marker.symbol`
      value, or `'cone'`.

  Returns:
    The list of plotly traces
  """
  lines_xyz_kwargs = make_lines_kwargs(
      start=start,
      end=end,
      axis=axis,
      end_marker=end_marker,
  )
  lines_trace = go.Scatter3d(**lines_xyz_kwargs)
  traces = [lines_trace]
  if end_marker is None or end_marker in _MARKER_SYMBOLS:
    pass
  elif end_marker == 'cone':
    cone_kwargs = make_cones_kwargs(
        start=start,
        direction=end - start,
        axis=axis,
    )
    cone_traces = go.Cone(
        **cone_kwargs,
        showlegend=False,
        showscale=False,
        # TODO(plotly): Absolute size currently broken:
        # https://github.com/plotly/plotly.js/issues/3613
        sizemode='absolute',  # Not sure what's the difference with `scaled`
        sizeref=0.5,
        # TODO(epot): Add color
        # colorscale=[[0, 'rgb(255,0,0)'], [1, 'rgb(255,0,0)']]
    )
    traces.append(cone_traces)
  else:
    raise ValueError(f'Invalid end_marker={end_marker!r}')
  return traces


def make_lines_kwargs(
    start: FloatArray['... 3'],
    end: FloatArray['... 3'],
    *,
    axis: int = -1,
    end_marker: Optional[str] = None,
) -> _PlotlyKwargs:
  """Returns the kwargs to plot lines."""
  assert axis == -1
  # 1) Flatten the arrays
  # Shape is `*d 3`
  assert start.shape == end.shape
  assert start.shape[-1] == 3

  start = start.reshape((-1, 3))
  end = end.reshape((-1, 3))

  # 2) Build the lines
  lines_xyz = [[], [], []]
  for s, e in zip(start, end):
    for i in range(3):  # (x, y, z)
      lines_xyz[i].append(s[i])
      lines_xyz[i].append(e[i])
      lines_xyz[i].append(None)  # line break

  # Add the marker kwargs
  if end_marker in _MARKER_SYMBOLS:
    marker_kwargs = dict(
        mode='lines+markers',
        marker=go.scatter3d.Marker(
            size=[0, 6, 0] * len(start),
            symbol=end_marker,
        ),
    )
  else:
    marker_kwargs = dict(mode='lines')

  return dict(
      **to_xyz_dict(lines_xyz, axis=0),
      **marker_kwargs,
  )


def make_cones_kwargs(
    start: FloatArray['... 3'],
    direction: FloatArray['... 3'],
    *,
    start_ratio: float = 0.98,
    axis: int = -1,
) -> _PlotlyKwargs:
  """Returns the kwargs to plot cones."""
  assert axis == -1
  # 1) Flatten the arrays
  # Shape is `*d 3`
  assert start.shape == direction.shape
  assert start.shape[-1] == 3

  start = start.reshape((-1, 3))
  direction = direction.reshape((-1, 3))

  # 2) Build the lines
  xyz = start + start_ratio * direction
  uvw = direction
  return {
      **to_xyz_dict(xyz),
      **to_xyz_dict(uvw, names='uvw'),
  }


def make_zero_point() -> plotly_base.BaseTraceType:
  """Returns the trace corresponding to the origin (0, 0, 0) point."""
  return go.Scatter3d(
      x=[0],
      y=[0],
      z=[0],
      showlegend=False,  # Do not display the legend
      marker=go.scatter3d.Marker(
          size=0.01,  # Small point to avoid accidental selection
          # Transparent color (somehow opacity=0. does not work)
          color='rgba(0, 0, 0, 0.0)',
      ),
  )


def _show_legend(traces: List[plotly_base.BaseTraceType]) -> bool:
  """Returns `False` if only a single trace has a legend."""
  # Should also group legend per group ?
  # Default: `t.showlegend is None` -> show legend
  return len([t for t in traces if t.showlegend is not False]) > 1  # pylint: disable=g-bool-id-comparison


def to_xyz_dict(
    arr: Array['... 3'],
    *,
    pattern: str = '{}',
    names: Union[str, Sequence[str]] = 'xyz',
    axis: int = -1,
) -> _PlotlyKwargs:
  """Convert np.array to xyz dict.

  Useful to create plotly kwargs from numpy arrays.

  Example:

  ```python
  to_xyz_dict(np.zeros((1, 3))) == {'x': [0], 'y': [0], 'z': [0]}
  to_xyz_dict(
    [0, 1, 2],
    pattern='axis_{}'
    names='uvw',
  ) == {'axis_u': 0, 'axis_v': 1, 'axis_w': 2}
  ```

  Args:
    arr: Array to convert
    pattern: Pattern to use for the axis names
    names: Names of the axis (default to 'x', 'y', 'z')
    axis: Axis containing the x, y, z coordinates to dispatch

  Returns:
    xyz_dict: The dict containing plotly kwargs.
  """
  arr = np.asarray(arr)
  if arr.shape[axis] != len(names):
    raise ValueError(f'Invalid shape: {arr.shape}[{axis}] != {len(names)}')

  # Build the `dict(x=arr[..., 0], y=arr[..., 1], z=arr[..., 2])`
  vals = {
      pattern.format(axis_name): arr_slice.flatten()
      for axis_name, arr_slice in zip(names, np.moveaxis(arr, axis, 0))
  }
  # Normalize scalars (as plotly reject `np.array(1)`)
  vals = {k: v if v.shape else v.item() for k, v in vals.items()}
  return vals


def subsample(
    *arrays: Optional[Array['... d']],
    num_samples: Optional[int],
) -> list[Optional[Array['...']]]:
  """Flatten and subsample the arrays (keeping the last dimension)."""
  if num_samples is None or num_samples == -1:  # No sampling
    return list(arrays)
  assert arrays[0] is not None
  shape = arrays[0].shape
  assert len(shape) >= 1
  batch_size = dca.utils.np_utils.size_of(shape[:-1])

  if batch_size > num_samples:
    # All arrays are sub-sampled the same way, so generate ids separately
    rng = np.random.default_rng(0)
    idx = rng.choice(batch_size, size=num_samples, replace=False)

  arrays_out = []
  for arr in arrays:
    if arr is None:
      arrays_out.append(None)
      continue
    if arr.shape[:-1] != shape[:-1]:
      raise ValueError('Incompatible shape')
    arr = einops.rearrange(arr, '... d -> (...) d')  # Flatten
    if batch_size > num_samples:
      arr = arr[idx]
    arrays_out.append(arr)

  return arrays_out


# Might want to extand this to PlotType.PLOT_2D, PlotType.PLOT_3D,... ?
def _is_traces_2d(traces: list[plotly_base.BaseTraceType]) -> bool:
  """Returns `True` if the traces are a 2d plot."""
  traces_2d_cls = (  # TODO(epot): Complete all the plotly types
      go.Scatter,
      go.Scattergl,
  )
  has_2d = False
  has_3d = False
  for trace in traces:
    if isinstance(trace, traces_2d_cls):
      has_2d = True
    else:
      has_3d = True
  if has_2d and has_3d:
    cls_names = {type(trace).__name__ for trace in traces}
    raise ValueError(
        f'Trying to mix 2d and 3d plots: {cls_names}. Please open a bug if '
        'this is an issue.'
    )
  return has_2d


def is_visualizable(item: VisualizableItem) -> bool:
  """Returns `True` if the element is a visualizable item."""
  return py_utils.supports_protocol(item, 'make_traces')


def has_fig_config(item: VisualizableItem) -> bool:
  return isinstance(
      getattr(item, 'fig_config', None), fig_config_utils.TraceConfig
  )



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interpolation utils."""

from etils import enp
from etils.array_types import Array, FloatArray  # pylint: disable=g-multiple-import


@enp.check_and_normalize_arrays(strict=False)
def interp_img(
    img: Array['h w c'],
    coords: FloatArray['*shape 2'],
    *,
    xnp: enp.NpModule = ...,
) -> FloatArray['*shape c']:
  """Bilinear interpolation of coordinates from an image.

  The following assumptions are made:

  * Interpolation is performed from the pixel center (`coords=[0.5, 0.5]`
    match `img[0, 0]`).
  * Pixel coordinates are given in `coords=[w, h]`, NOT `[h, w]`. This is
    consistent with `cam.spec.px_centers()` and OpenCv conventions.
    To invert, use `coords[..., ::-1]`
  * Coordinates outside the images use the image corners value.

  Args:
    img: Image from which interpolating values
    coords: Pixel coordinates from which interpolate (in `w, h` convention).
    xnp: Numpy module to use

  Returns:
    The interpolated pixel values at the requested coordinates.
    Note that returned value is always float (so interpolation of `uint8`
    return float)
  """
  h, w, c = img.shape
  *coord_shape, _ = coords.shape

  # From https://en.wikipedia.org/wiki/Bilinear_interpolation

  # TODO(b/262392450): Should use `.flatten()` instead
  x = coords[..., 0].reshape((-1,))  # w
  y = coords[..., 1].reshape((-1,))  # h

  # For each query coordinates, extract the 4 corners coordinates
  # Pixel coordinates are centered, so keep both `_i` (integer index) and
  # `_f` (float centered coordinates).
  x0 = xnp.floor(x - 0.5)
  x0f = x0 + 0.5
  x0i = enp.compat.astype(x0, xnp.int32)
  x1f = x0f + 1
  x1i = x0i + 1
  y0 = enp.compat.astype(xnp.floor(y - 0.5), xnp.int32)
  y0f = y0 + 0.5
  y0i = enp.compat.astype(y0, xnp.int32)
  y1f = y0f + 1
  y1i = y0i + 1

  del x0, y0  # Should explicitly use integer `_i` or float `_f`

  # Compute the weights for the 4 corners
  # This has to be applied before clipping
  wa = (x1f - x) * (y1f - y)
  wb = (x1f - x) * (y - y0f)
  wc = (x - x0f) * (y1f - y)
  wd = (x - x0f) * (y - y0f)

  # Clip coordinates outside the image
  x0i = xnp.clip(x0i, 0, w - 1)
  x1i = xnp.clip(x1i, 0, w - 1)
  y0i = xnp.clip(y0i, 0, h - 1)
  y1i = xnp.clip(y1i, 0, h - 1)

  if enp.lazy.is_torch_xnp(xnp):
    # Pytorch indexing do not support int32
    x0i = enp.compat.astype(x0i, xnp.int64)
    x1i = enp.compat.astype(x1i, xnp.int64)
    y0i = enp.compat.astype(y0i, xnp.int64)
    y1i = enp.compat.astype(y1i, xnp.int64)

  # Extract the `(num_coords, c)` value for each corners
  val_a = img[y0i, x0i]
  val_b = img[y1i, x0i]
  val_c = img[y0i, x1i]
  val_d = img[y1i, x1i]

  # Weighted sum for each corners
  # `(num_coords, c) * (num_coords, 1) -> (num_coords, c)`
  pts = (
      val_a * wa[..., None]
      + val_b * wb[..., None]
      + val_c * wc[..., None]
      + val_d * wd[..., None]
  )

  assert pts.shape[-1] == c
  return pts.reshape(coord_shape + [c])



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataclass array wrapper."""

from __future__ import annotations

import dataclasses
from typing import Any

import dataclass_array as dca
from dataclass_array.typing import DcT
from visu3d.plotly import fig_config_utils
from visu3d.plotly import fig_utils


@dca.dataclass_array(broadcast=True, cast_dtype=True)
class DataclassArray(dca.DataclassArray, fig_utils.Visualizable):
  """Wrapper around `dca.DataclassArray` for all v3d objects.

  This class is like `dca.DataclassArray` but in addition:

  * Add the `my_obj.fig` property to all objects.
  * Add a `my_obj.fig_config` property to control object display options. As
    `dca.DataclassArray` are immutable, options can be updated using
    `my_obj = my_obj.replace_fig_config(**options)`.

  """

  __dca_non_init_fields__ = ('fig_config',)

  # Note: Because `FigConfig` is immutable, it is safe to use a shared instance
  # to avoid unecessary copy.
  fig_config: fig_config_utils.TraceConfig = dataclasses.field(
      default=fig_config_utils.TraceConfig(),
      repr=False,
      init=False,
  )

  def replace_fig_config(
      self: DcT,
      *,
      name: str = ...,  # pytype: disable=annotation-type-mismatch
      num_samples: int = ...,  # pytype: disable=annotation-type-mismatch
      **kwargs: Any,
  ) -> DcT:
    """Returns a copy of self with figure params overwritten."""
    fig_config_kwargs = dict(
        name=name,
        num_samples=num_samples,
        **kwargs,
    )
    # Filter Ellipsis values
    fig_config_kwargs = {
        k: v for k, v in fig_config_kwargs.items() if v is not ...
    }
    return self.replace(fig_config=self.fig_config.replace(**fig_config_kwargs))



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Camera util."""

from __future__ import annotations

import dataclasses
from typing import Any

import dataclass_array as dca
from dataclass_array.typing import DcT
from etils import enp
from etils.array_types import FloatArray  # pylint: disable=g-multiple-import
import numpy as np
from visu3d import array_dataclass
from visu3d import plotly
from visu3d.dc_arrays import camera_spec
from visu3d.dc_arrays import point as point_lib
from visu3d.dc_arrays import ray as ray_lib
from visu3d.dc_arrays import transformation
from visu3d.utils import np_utils
from visu3d.utils.lazy_imports import plotly_base


class Camera(array_dataclass.DataclassArray):
  """A camera located in space.

  Attributes:
    spec: Camera intrinsics parameters
    world_from_cam: Camera pose (`v3d.Transformation`)
  """

  spec: camera_spec.CameraSpec
  world_from_cam: transformation.Transform

  # Overwrite `v3d.DataclassArray.fig_config`.
  fig_config: camera_spec.TraceConfig = dataclasses.field(
      default=camera_spec.TraceConfig(),
      repr=False,
      init=False,
  )

  @classmethod
  def from_look_at(
      cls,
      *,
      pos: FloatArray['*shape 3'],
      target: FloatArray['*shape 3'],
      spec: camera_spec.CameraSpec,
  ) -> Camera:
    """Factory which creates a camera looking at `target`.

    This assume the camera is parallel to the floor. See `v3d.CameraSpec`
    for axis conventions.

    Args:
      pos: Origin position
      target: Target position
      spec: Camera specifications.

    Returns:
      cam: Camera pointing to the ray.
    """
    world_from_cam = transformation.Transform.from_look_at(
        target=target,
        pos=pos,
    )
    return cls(spec=spec, world_from_cam=world_from_cam)

  @dca.vectorize_method
  def look_at(self, target: FloatArray['*shape 3']) -> Camera:
    """Returns a new camera looking at the target."""
    world_from_cam = self.world_from_cam.look_at(target)
    return self.replace(world_from_cam=world_from_cam)

  @property
  def resolution(self) -> tuple[int, int]:
    """`(h, w)` resolution in pixel."""
    return self.spec.resolution

  @property
  def hw(self) -> tuple[int, int]:
    """`(Height, Width)` in pixel (for usage in `(i, j)` coordinates)."""
    return self.spec.hw

  @property
  def wh(self) -> tuple[int, int]:
    """`(Width, Height)` in pixel (for usage in `(u, v)` coordinates)."""
    return self.spec.wh

  @property
  def h(self) -> int:
    """Height in pixel."""
    return self.spec.h

  @property
  def w(self) -> int:
    """Width in pixel."""
    return self.spec.w

  def __add__(self, translation: FloatArray['... 3']) -> Camera:
    """Translate the position."""
    return self.replace(world_from_cam=self.world_from_cam + translation)

  __sub__ = np_utils.__sub__

  @dca.vectorize_method(static_args={'normalize'})
  def rays(self, normalize: bool = True) -> ray_lib.Ray:
    """Creates the rays.

    Args:
      normalize: If `False`, returns camera rays in the `z=1` from the camera
        frame.

    Returns:
      rays: Pose
    """
    cam_dir = self.spec.cam_from_px @ self.spec.px_centers()
    # Rotate the points from cam -> world
    world_dir = self.world_from_cam.apply_to_dir(cam_dir)
    # Position is (0, 0, 0) so no need to transform

    rays = ray_lib.Ray(pos=self.world_from_cam.t, dir=world_dir)
    if normalize:
      rays = rays.normalize()
    return rays

  @property
  def cam_from_world(self) -> transformation.Transform:
    """Transformation from 3d world coordinates to 3d camera coordinates."""
    return self.world_from_cam.inv

  @property
  def px_from_world(self) -> transformation.TransformBase:
    """Transfomration from 3d world coordinates to 2d pixel coordinates."""
    return self.spec.px_from_cam @ self.cam_from_world

  @property
  def world_from_px(self) -> transformation.TransformBase:
    """Transfomration from 2d pixel coordinates to 3d world coordinates."""
    return self.world_from_cam @ self.spec.cam_from_px

  @dca.vectorize_method
  def render(self, points: point_lib.Point3d) -> FloatArray['*shape h w 3']:
    """Project 3d points to the camera screen.

    Args:
      points: 3d points.

    Returns:
      img: The projected 3d points.
    """
    # TODO(epot): Support float colors and make this differentiable!
    if not isinstance(points, point_lib.Point3d):
      raise TypeError(
          f'Camera.render expect `v3d.Point3d` as input. Got: {points}.'
      )

    # Project 3d -> 2d coordinates
    points2d = self.px_from_world @ points

    # Flatten pixels
    points2d = points2d.flatten()
    px_coords = points2d.p
    rgb = points2d.rgb

    # Compute the valid coordinates
    w_coords = px_coords[..., 0]
    h_coords = px_coords[..., 1]
    valid_coords_mask = (
        (0 <= h_coords)
        & (h_coords < self.h - 1)
        & (0 <= w_coords)
        & (w_coords < self.w - 1)
        & (points2d.depth[..., 0] > 0)  # Filter points behind the camera
    )
    rgb = rgb[valid_coords_mask]
    px_coords = px_coords[valid_coords_mask]
    px_coords = enp.compat.astype(enp.compat.round(px_coords), self.xnp.int32)

    # TODO(epot): Should we create a `xnp.asarray` ?
    # TODO(epot): The dtype should be cloned from point.rgb !
    img = np.zeros((*self.resolution, 3), dtype=np.uint8)
    # px_coords is (h, w)
    img[px_coords[..., 1], px_coords[..., 0]] = rgb
    return img

  # Could be removed but only kept for type-checking / auto-complete.
  def replace_fig_config(  # pylint: disable=useless-parent-delegation
      self: DcT,
      *,
      name: str = ...,  # pytype: disable=annotation-type-mismatch
      scale: float = ...,  # pytype: disable=annotation-type-mismatch
      **kwargs: Any,
  ) -> DcT:
    """Returns a copy of self with figure params overwritten."""
    return super().replace_fig_config(name=name, scale=scale, **kwargs)

  # Protocols (inherited)

  def apply_transform(self, tr: transformation.Transform) -> Camera:
    return self.replace(world_from_cam=tr @ self.world_from_cam)

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    spec = self.spec.replace(fig_config=self.fig_config)
    # TODO(epot): Add arrow to indicates the orientation ?
    start, end = spec._get_camera_lines()  # pylint: disable=protected-access
    start = self.world_from_cam @ start
    end = self.world_from_cam @ end
    return plotly.make_lines_traces(start=start, end=end)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for camera."""

from __future__ import annotations

import dataclass_array as dca
from etils import enp
import numpy as np
import pytest
import visu3d as v3d

# Activate the fixture
enable_tf_np_mode = enp.testing.set_tnp


H, W = 16, 32


parametrize_spec = pytest.mark.parametrize(
    'spec',
    [
        v3d.PinholeCamera,
        v3d.Spec360,
    ],
)


def _make_spec(spec: type[v3d.CameraSpec]) -> v3d.CameraSpec:
  if spec is v3d.PinholeCamera:
    return v3d.PinholeCamera.from_focal(resolution=(H, W), focal_in_px=35.0)
  elif spec is v3d.Spec360:
    return v3d.Spec360(resolution=(H, W))
  else:
    raise AssertionError(spec)


def _make_cam(
    *,
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
    spec: type[v3d.CameraSpec],
) -> v3d.Camera:
  """Create a camera at (0, 4, 0) looking at the center."""
  spec = _make_spec(spec)
  cam = v3d.Camera.from_look_at(
      spec=spec.as_xnp(xnp),
      pos=[0, 4, 0],  # Camera on the `y` axis
      target=[0, 0, 0],
  )
  np.testing.assert_allclose(cam.world_from_cam.scale_xyz, [1, 1, 1])
  cam = cam.broadcast_to(shape)
  return cam


@enp.testing.parametrize_xnp()
@parametrize_spec
def test_camera_xnp(xnp: enp.NpModule, spec: type[v3d.CameraSpec]):
  spec = _make_spec(spec)

  assert (
      v3d.Camera.from_look_at(
          spec=spec,
          pos=[0, 4.0, 0],  # Camera on the `y` axis
          target=[0, 0.0, 0],
      ).xnp
      is np
  )
  assert (
      v3d.Camera.from_look_at(
          spec=spec.as_xnp(xnp),
          pos=[0, 4.0, 0],
          target=[0, 0.0, 0],
      ).xnp
      is xnp
  )
  assert (
      v3d.Camera.from_look_at(
          spec=spec,
          pos=xnp.asarray([0, 4.0, 0]),
          target=[0, 0.0, 0],
      ).xnp
      is xnp
  )
  assert (
      v3d.Camera.from_look_at(
          spec=spec,
          pos=[0, 4.0, 0],
          target=xnp.asarray([0.0, 0, 0]),
      ).xnp
      is xnp
  )


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('shape', [(), (3,)])
@parametrize_spec
def test_camera_properties(
    xnp: enp.NpModule, shape: dca.typing.Shape, spec: type[v3d.CameraSpec]
):
  cam = _make_cam(xnp=xnp, shape=shape, spec=spec)

  # Properties
  assert cam.shape == shape
  assert cam.spec.shape == shape
  assert cam.world_from_cam.shape == shape
  assert cam.resolution == (H, W)
  assert cam.h == H
  assert cam.w == W
  assert cam.wh == (W, H)
  assert cam.hw == (H, W)


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('shape', [(), (3,), (1, 1)])
@pytest.mark.parametrize('normalize', [False, True])
@parametrize_spec
def test_camera_rays(
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
    normalize: bool,
    spec: type[v3d.CameraSpec],
):
  if shape and xnp in [
      enp.lazy.tnp,
  ]:
    pytest.skip('Vectorization not supported yet with TF')

  cam = _make_cam(xnp=xnp, shape=shape, spec=spec)

  rays = cam.rays(normalize=normalize)
  assert isinstance(rays, v3d.Ray)
  assert rays.xnp is xnp
  assert rays.shape == shape + (H, W)
  np.testing.assert_allclose(
      rays.pos,
      cam.world_from_cam[..., None, None].broadcast_to(shape + (H, W)).t,
  )

  if normalize or spec == v3d.Spec360:
    np.testing.assert_allclose(rays.norm(), np.ones(shape + (H, W)), atol=1e-6)
    # Ray is normalized
    np.testing.assert_allclose(
        np.linalg.norm(rays.dir, axis=-1),
        np.ones(shape + (H, W)),
        atol=1e-6,
    )
  elif spec == v3d.PinholeCamera:
    # Ray destinations are aligned with the y=3 plane
    np.testing.assert_allclose(rays.end[..., 1], np.full(shape + (H, W), 3.0))
  else:
    raise AssertionError

  dca.testing.assert_array_equal(cam + [0, 0, 0], cam)
  dca.testing.assert_array_equal(cam + xnp.asarray([0, 0, 0]), cam)
  dca.testing.assert_array_equal(cam - xnp.asarray([0, 0, 0]), cam)

  _ = cam.fig

  cam = cam.replace_fig_config(scale=3.0)
  assert cam.fig_config.scale == 3.0
  _ = cam.fig


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('shape', [(), (3,), (2, 2)])
@parametrize_spec
def test_camera_transform(
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
    spec: type[v3d.CameraSpec],
):
  """Test low-level camera transforms."""

  if shape and xnp in [
      enp.lazy.tnp,
      # Batching rule not implemented for aten::concatenate
      enp.lazy.torch,
  ]:
    pytest.skip('Vectorization not supported yet with TF')

  cam = _make_cam(xnp=xnp, shape=shape, spec=spec)  # Camera on the `y` axis
  rays = cam.rays(normalize=False)

  with enp.lazy.jax.checking_leaks():
    # px -> cam -> world coordinates
    px = cam.spec.px_centers()
    pts0 = cam.world_from_cam @ (cam.spec.cam_from_px @ px)
    pts1 = (cam.world_from_cam @ cam.spec.cam_from_px) @ px
    pts2 = cam.world_from_px @ px
    np.testing.assert_allclose(pts0, pts1, atol=1e-5, rtol=1e-6)
    np.testing.assert_allclose(pts0, pts2, atol=1e-5, rtol=1e-6)
    np.testing.assert_allclose(rays.end, pts0, atol=1e-5, rtol=1e-6)

    # world -> cam -> px coordinates
    px0 = cam.spec.px_from_cam @ (cam.cam_from_world @ pts0)
    px1 = (cam.spec.px_from_cam @ cam.cam_from_world) @ pts0
    px2 = cam.px_from_world @ pts0
    np.testing.assert_allclose(px0, px, atol=1e-5, rtol=1e-6)
    np.testing.assert_allclose(px1, px, atol=1e-5, rtol=1e-6)
    np.testing.assert_allclose(px2, px, atol=1e-5, rtol=1e-6)


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize('shape', [(), (3,), (1, 1)])
@pytest.mark.parametrize('point_shape', [(), (3,), (1, 1)])
@parametrize_spec
def test_camera_render(
    xnp: enp.NpModule,
    shape: dca.typing.Shape,
    point_shape: dca.typing.Shape,
    spec: type[v3d.CameraSpec],
):
  if shape and xnp in [
      enp.lazy.tnp,
      # Batching rule not implemented for aten::concatenate.
      enp.lazy.torch,
  ]:
    pytest.skip('Vectorization not supported yet with TF')

  cam = _make_cam(xnp=xnp, shape=shape, spec=spec)  # Camera on the `y` axis
  points = v3d.Point3d(
      p=[0, 0, 0],
      rgb=[255.0, 255.0, 255.0],
  )
  points = points.as_xnp(xnp)
  points = points.broadcast_to((1,) * len(shape) + point_shape)
  if shape and xnp is enp.lazy.jnp:
    pytest.skip('vmap not (yet) supported for cam.render')
  img = cam.render(points)
  assert img.shape == shape + (H, W, 3)
  assert img.dtype == np.uint8
  first_idx = (0,) * len(shape)
  np.testing.assert_allclose(img[(*first_idx, 0, 0)], [0, 0, 0])
  np.testing.assert_allclose(img[(*first_idx, H // 2, W // 2)], [255, 255, 255])



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Math and geometry API."""

from visu3d.math.coord_utils import carthesian_to_spherical
from visu3d.math.coord_utils import spherical_to_carthesian
from visu3d.math.interp_utils import interp_img
from visu3d.math.math_utils import subsample
from visu3d.math.rotation_utils import DEG2RAD
from visu3d.math.rotation_utils import euler_to_rot
from visu3d.math.rotation_utils import is_orth
from visu3d.math.rotation_utils import is_rot
from visu3d.math.rotation_utils import RAD2DEG
from visu3d.math.rotation_utils import rot_to_euler
from visu3d.math.rotation_utils import rot_x
from visu3d.math.rotation_utils import rot_y
from visu3d.math.rotation_utils import rot_z
from visu3d.utils.np_utils import interp_points



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotly API."""

from __future__ import annotations

from visu3d.plotly.fig_config_utils import FigConfig
from visu3d.plotly.fig_config_utils import TraceConfig
from visu3d.plotly.fig_utils import make_cones_kwargs
from visu3d.plotly.fig_utils import make_fig
from visu3d.plotly.fig_utils import make_lines_kwargs
from visu3d.plotly.fig_utils import make_lines_traces
from visu3d.plotly.fig_utils import make_points
from visu3d.plotly.fig_utils import make_traces
from visu3d.plotly.fig_utils import make_zero_point
from visu3d.plotly.fig_utils import subsample
from visu3d.plotly.fig_utils import to_xyz_dict
from visu3d.plotly.fig_utils import Visualizable



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test."""

from etils import enp
import numpy as np
import visu3d as v3d

# Activate the fixture
enable_tf_np_mode = enp.testing.set_tnp


@enp.testing.parametrize_xnp()
def test_interp_img(xnp: enp.NpModule):
  img = xnp.asarray([
      [0, 1, 2, 3],
      [1, 11, 21, 31],
      [2, 12, 22, 32],
  ])
  out = v3d.math.interp_img(
      img[..., None],
      coords=[
          [0.5, 0.5],  # Top-left corner
          [2.5, 1.5],  # Inside
          [2.0, 0.5],  # Between 2 cells
          [5.0, 1.5],  # Out of boundaries
      ],
  )
  assert out.shape == (4, 1)
  assert enp.lazy.get_xnp(out) is xnp

  np.testing.assert_allclose(
      out,
      [
          [0],
          [21],
          [1.5],
          [31],
      ],
      atol=1e-6,
  )


@enp.testing.parametrize_xnp()
def test_interp_img_center(xnp: enp.NpModule):
  # Interpolating an image using the pixel coordinates is a no-op
  h, w = 12, 30
  spec = v3d.PinholeCamera.from_focal(resolution=(h, w), focal_in_px=34.0)
  spec = spec.as_xnp(xnp)

  rng = np.random.default_rng(0)
  img = rng.random((h, w, 3))
  centers = spec.px_centers()
  out = v3d.math.interp_img(img, coords=centers)
  assert out.shape == (h, w, 3)
  assert enp.lazy.get_xnp(out) is xnp

  # Sampling the pixel center should be a no-op
  np.testing.assert_allclose(out, img, atol=1e-6)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Numpy utils.

And utils intended to work on both `xnp.ndarray` and `v3d.DataclassArray`.
"""

from __future__ import annotations

from typing import TypeVar, Union

import dataclass_array as dca
import einops
from etils import enp
from etils.array_types import FloatArray  # pylint: disable=g-multiple-import

_T = TypeVar('_T')

# Maybe some of those could live in `enp` ?


def append_row(
    x: FloatArray['*d daxis'],
    value: float,
    *,
    axis: int,  # Axis is required as `row` imply `axis=0` while we want `=-1`
) -> FloatArray['*d daxis+1']:
  """Like `np.append`, but broadcast the value to `x` shape."""
  xnp = enp.get_np_module(x)
  value = xnp.asarray(value)
  if len(value.shape) == 0:  # pylint: disable=g-explicit-length-test
    shape = list(x.shape)
    shape[axis] = 1
    value = xnp.broadcast_to(value, shape)
  elif len(value.shape) == 1:
    # TODO(epot): support actual row: append_row(x, [0, 0, 0, 1]). Might require
    # adding a `broadcast_to` which support arbitrary array.
    assert x.shape[axis] == len(value)
    raise NotImplementedError()
  else:
    raise ValueError(
        f'`append_row` does not support appending rank > 1. Got {value.shape}.'
    )
  return enp.compat.concat([x, value], axis=axis)


def interp_points(
    points: FloatArray['num_points d'],
    *,
    t: Union[int, FloatArray['t']],
    axis: int = -1,
    **splprep_kwargs,
) -> FloatArray['t d']:
  """Spline interpolation between x-d Points.

  Args:
    points: Key points of shape `(num_point, num_dims)`
    t: Either number of steps (in which case steps are linearly interpolated),
      or an array of `[0, 1]` values corresponding to the)
    axis: Axis
    **splprep_kwargs: Kwargs forwarded to `scipy.interpolate.splprep`

  Returns:
    The interpolated points of shape `(t, num_dims)`
  """
  import scipy.interpolate  # pylint: disable=g-import-not-at-top

  if isinstance(points, dca.DataclassArray):
    # Could eventually add a protocol to support interpolation between Camera,
    # Ray,...
    raise NotImplementedError('DataclassArray not supported. Only `np.array`.')
  xnp = enp.lazy.get_xnp(points, strict=False)
  # TODO(epot): Currently only np is supported
  if xnp is not enp.lazy.np:
    raise NotImplementedError(f'Only numpy supported. Not {xnp}')
  if axis != -1:
    raise NotImplementedError(
        'interpolation currently only supports axis=-1. Open an issue if you '
        'need this.'
    )
  points = xnp.asarray(points)
  points = einops.rearrange(points, 'num_points d -> d num_points')

  tck, _ = scipy.interpolate.splprep(points, **splprep_kwargs)

  if isinstance(t, int):
    t = xnp.linspace(0, 1, t)
  points_fine = scipy.interpolate.splev(t, tck)
  points_fine = einops.rearrange(points_fine, 'd t -> t d')
  return points_fine


def __sub__(self: _T, translation: FloatArray['... 3']) -> _T:  # pylint: disable=invalid-name
  """Add `my_obj - array` support, assuming `my_obj + array` exists."""
  translation = self.xnp.asarray(translation)
  return self + (-translation)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Math utils."""

from typing import Optional

import dataclass_array as dca
from dataclass_array.typing import DcT
import numpy as np


def subsample(
    array: DcT,
    *,
    num_samples: Optional[int],
    seed: Optional[int] = None,
) -> DcT:
  """Subsample the given DataclassArray."""
  if not isinstance(array, dca.DataclassArray):
    raise TypeError(f'Subsample only accept DataclassArray. Got: {type(array)}')

  if num_samples in (None, -1) or array.size <= num_samples:
    return array  # No-op

  rng = np.random.default_rng(seed)
  idx = rng.choice(array.size, size=num_samples, replace=False)

  array = array.flatten()
  array = array[idx]
  return array



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visu3d API."""

from __future__ import annotations

# pylint: disable=g-import-not-at-top,g-bad-import-order

# Core API
from visu3d.array_dataclass import DataclassArray
from visu3d.utils import lazy_imports

from visu3d import math
from visu3d.math import DEG2RAD
from visu3d.math import RAD2DEG

# Visualization
from visu3d import plotly
from visu3d.plotly import make_fig
from visu3d.plotly import make_traces
from visu3d.plotly import Visualizable
from visu3d.plotly.auto_plot import auto_plot_figs
from visu3d.plotly.fig_config_utils import fig_config

# Arrays implementation
from visu3d.dc_arrays.camera import Camera
from visu3d.dc_arrays.camera_spec import CameraSpec
from visu3d.dc_arrays.camera_spec import PinholeCamera
from visu3d.dc_arrays.camera_spec import Spec360
from visu3d.dc_arrays.point import Point3d
from visu3d.dc_arrays.point import Point2d
from visu3d.dc_arrays.ray import Ray
from visu3d.dc_arrays.transformation import custom_transform
from visu3d.dc_arrays.transformation import Transform

# Updating this will auto-trigger a release on PyPI and GitHub
# Note:
# * Make sure to also update the `CHANGELOG.md` before this.
# * Make sure to also trigger an `etils` release
__version__ = '1.5.0'



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Auto plot utils."""

from __future__ import annotations

from visu3d.plotly import fig_utils
from visu3d.utils.lazy_imports import IPython


def auto_plot_figs() -> None:
  """Auto-display `tuple[Visualizable, ...]` as figure.

  After this function is called, tuple of `v3d.Visualizable` objects are
  displayed directly as if they were passed to `v3d.make_fig`.

  >>> rays, cams, points

  Is an alias for

  >>> v3d.make_fig([rays, cams, points])

  It uses a simple heuristic (only check is the first element is visualizable).
  Other tuple are display as-is.

  """
  ipython = IPython.get_ipython()
  if ipython is None:
    return  # Non-notebook environement

  def make_fig(val):
    # Only check the first 5 elements
    if not val or not any(fig_utils.is_visualizable(v) for v in val[:5]):
      return None
    fig_utils.make_fig(list(val))._ipython_display_()  # pylint: disable=protected-access
    return ''

  print('Display `tuple[v3d.Visualizable, ...]` as figure')
  formatter = ipython.display_formatter.formatters['text/html']
  formatter.for_type(tuple, make_fig)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Coordinate conversion utils."""

from typing import NamedTuple, Optional

from dataclass_array.typing import FloatArray
from etils import enp


class _SphericalCoords(NamedTuple):
  """(r, theta, phi) == (r, azimuth, elevation).

  Follow https://mathworld.wolfram.com/SphericalCoordinates.html
  conventions.

  Attributes:
    r: radial
    theta: azimuth / longitude `(0, tau)`
    elevation: polar / elevation /  colatitude `(0, tau/2)`
  """

  r: FloatArray['*shape']
  theta: FloatArray['*shape']
  phi: FloatArray['*shape']


@enp.check_and_normalize_arrays(strict=False)
def carthesian_to_spherical(
    point3d: FloatArray['... 3'],
    *,
    xnp: enp.NpModule = ...,
) -> _SphericalCoords:
  """Convert (x, y, z) to (r, theta, phi).

  Follow https://mathworld.wolfram.com/SphericalCoordinates.html
  conventions.

  Args:
    point3d: 3d carthesian coordinates
    xnp: Numpy module

  Returns:
    r: radial
    theta: azimuth / longitude in range `(0, tau)`
    elevation: polar / elevation /  colatitude in range `(0, tau/2)`
  """
  r = enp.compat.norm(point3d, axis=-1)
  theta = xnp.arctan2(point3d[..., 1], point3d[..., 0])  # (-tau/2, tau/2)
  theta = theta % enp.tau  # Normalize azimuth (-tau/2, tau/2) -> (0, tau)
  phi = xnp.arccos(point3d[..., 2] / r)  # elevation (0, tau/2)

  return _SphericalCoords(r, theta, phi)


@enp.check_and_normalize_arrays(strict=False)
def spherical_to_carthesian(
    r: Optional[FloatArray['*shape']] = None,
    theta: Optional[FloatArray['*shape']] = None,
    phi: Optional[FloatArray['*shape']] = None,
    *,
    xnp: enp.NpModule = ...,
) -> FloatArray['*shape 3']:
  """Convert (r, theta, phi) to (x, y, z).

  Follow https://mathworld.wolfram.com/SphericalCoordinates.html
  conventions.

  Usage:

  ```
  v3d.math.spherical_to_carthesian(r, theta, phi)
  v3d.math.spherical_to_carthesian(theta=theta, phi=phi)
  ```

  Args:
    r: If `None`, assume 1.
    theta: azimuth / longitude
    phi: polar / elevation /  colatitude
    xnp: Numpy module

  Returns:
    points: 3d coordinates
  """
  if theta is None or phi is None:
    raise ValueError('theta and phi should be given.')
  points3d = xnp.stack(
      [
          xnp.sin(phi) * xnp.cos(theta),
          xnp.sin(phi) * xnp.sin(theta),
          xnp.cos(phi),
      ],
      axis=-1,
  )
  if r is not None:
    points3d = points3d * r[..., None]
  return points3d



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for transformation."""

from __future__ import annotations

import dataclasses
import functools
from typing import Optional

import dataclass_array as dca
from etils import enp
from etils.array_types import FloatArray
import numpy as np
import pytest
import visu3d as v3d

# Activate the fixture
enable_tf_np_mode = enp.testing.set_tnp


@dataclasses.dataclass
class TransformExpectedValue:
  """Tests values."""

  # Expected rays values after transformation
  expected_pos: FloatArray[..., 3]
  expected_dir: FloatArray[..., 3]

  # Expected transformation values after composition with other tr
  expected_r: FloatArray[..., 3, 3]
  expected_t: FloatArray[..., 3]

  # Transformation params
  R: Optional[FloatArray[3, 3]] = None  # pylint: disable=invalid-name
  t: Optional[FloatArray[3]] = None


# Transformation values
_RAY_POS = np.array([1, 3, 5], dtype=np.float32)
_RAY_DIR = np.array([2, 1, 4], dtype=np.float32)
_TR_R = np.array(
    [
        [0, 1, 0],
        [1, 0, 0],
        [0, 0, 1],
    ],
    dtype=np.float32,
)
_TR_T = np.array([4, 3, 7], dtype=np.float32)

_TR_EXPECTED_VALUES = {
    # Test identity
    'default_args': TransformExpectedValue(
        # Identity should be a no-op
        expected_pos=_RAY_POS,
        expected_dir=_RAY_DIR,
        expected_r=_TR_R,
        expected_t=_TR_T,
    ),
    # Test identity
    'identity': TransformExpectedValue(
        R=np.eye(3, dtype=np.float32),
        t=np.zeros((3,), dtype=np.float32),
        # Identity should be a no-op
        expected_pos=_RAY_POS,
        expected_dir=_RAY_DIR,
        expected_r=_TR_R,
        expected_t=_TR_T,
    ),
    # Test translation only
    'trans_only': TransformExpectedValue(
        R=np.eye(3, dtype=np.float32),
        t=[3, -1, 2],
        # Only position translated
        expected_pos=_RAY_POS + [3, -1, 2],
        expected_dir=_RAY_DIR,
        expected_r=_TR_R,
        expected_t=_TR_T + [3, -1, 2],
    ),
    # Test translation only
    'trans_only_default_arg': TransformExpectedValue(
        t=[3, -1, 2],
        # Only position translated
        expected_pos=_RAY_POS + [3, -1, 2],
        expected_dir=_RAY_DIR,
        expected_r=_TR_R,
        expected_t=_TR_T + [3, -1, 2],
    ),
    # Test rotation only
    'rot_only': TransformExpectedValue(
        R=[
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0],
        ],
        t=[0, 0, 0],
        # Rotation invert axis `(1, 2, 3)` -> `(3, 1, 2)`
        expected_pos=[5, 1, 3],
        expected_dir=[4, 2, 1],
        # Rotation invert axis `(1, 2, 3)` -> `(2, 1, 3)`
        # Rotation invert axis `(1, 2, 3)` -> `(3, 1, 2)`
        expected_r=[
            [0, 0, 1],
            [0, 1, 0],
            [1, 0, 0],
        ],
        expected_t=[7, 4, 3],
    ),
    # Test rotation only
    'rot_only_default_arg': TransformExpectedValue(
        R=[
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0],
        ],
        # Rotation invert axis `(1, 2, 3)` -> `(3, 1, 2)`
        expected_pos=[5, 1, 3],
        expected_dir=[4, 2, 1],
        # Rotation invert axis `(1, 2, 3)` -> `(2, 1, 3)`
        # Rotation invert axis `(1, 2, 3)` -> `(3, 1, 2)`
        expected_r=[
            [0, 0, 1],
            [0, 1, 0],
            [1, 0, 0],
        ],
        expected_t=[7, 4, 3],
    ),
    # Test translation + rotation
    'rot_trans': TransformExpectedValue(
        R=[
            [0, 0, 1],
            [1, 0, 0],
            [0, 1, 0],
        ],
        t=[3, -1, 2],
        # Rotation invert axis `(1, 2, 3)` -> `(3, 1, 2)`
        # And translation applied
        expected_pos=np.array([5, 1, 3]) + [3, -1, 2],
        expected_dir=[4, 2, 1],
        # Rotation invert axis `(1, 2, 3)` -> `(2, 1, 3)`
        # Rotation invert axis `(1, 2, 3)` -> `(3, 1, 2)`
        expected_r=[
            [0, 0, 1],
            [0, 1, 0],
            [1, 0, 0],
        ],
        expected_t=np.array([7, 4, 3]) + [3, -1, 2],
    ),
}


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize(
    'tr_shape, other_shape, expected_shape',
    [
        ((), (), ()),
        ((3,), (3,), (3,)),
        ((3,), (3, 4, 5), (3, 4, 5)),
        ((3,), (1, 4, 5), (3, 4, 5)),
        ((1,), (3, 4, 5), (3, 4, 5)),
    ],
)
@pytest.mark.parametrize(
    'test_values',
    _TR_EXPECTED_VALUES.values(),
    ids=_TR_EXPECTED_VALUES.keys(),
)
def test_transformation(
    xnp: enp.NpModule,
    tr_shape: dca.typing.Shape,
    other_shape: dca.typing.Shape,
    expected_shape: dca.typing.Shape,
    test_values: TransformExpectedValue,
):
  init_kwargs = {}
  if test_values.R is not None:
    init_kwargs['R'] = xnp.asarray(test_values.R)
  if test_values.t is not None:
    init_kwargs['t'] = xnp.asarray(test_values.t)
  tr = v3d.Transform(**init_kwargs)

  if test_values.R is None and test_values.t is None:
    tr = tr.as_xnp(xnp)
  tr = tr.broadcast_to(tr_shape)

  if tr_shape and xnp in [
      enp.lazy.tnp,
      # Batching rule not implemented for aten::concatenate.
      enp.lazy.torch,
  ]:
    pytest.skip('Vectorization not supported yet with TF')

  _assert_tr_common(tr, tr_shape=tr_shape)

  _assert_ray_transformed(
      tr,
      expected_shape=expected_shape,
      other_shape=other_shape,
      test_values=test_values,
  )

  _assert_point_transformed(
      tr,
      expected_shape=expected_shape,
      other_shape=other_shape,
      test_values=test_values,
  )

  _assert_tr_transformed(
      tr,
      expected_shape=expected_shape,
      other_shape=other_shape,
      test_values=test_values,
  )

  _assert_camera_transformed(
      tr,
      expected_shape=expected_shape,
      other_shape=other_shape,
      test_values=test_values,
  )


def _assert_tr_common(tr: v3d.Transform, tr_shape: dca.typing.Shape):
  """Generic rules applied to all transformation."""
  assert tr.shape == tr_shape
  assert tr.R.shape == tr_shape + (3, 3)
  assert tr.t.shape == tr_shape + (3,)

  identity_tr = v3d.Transform(R=tr.xnp.eye(3), t=tr.xnp.zeros((3,)))
  dca.testing.assert_array_equal(
      identity_tr,
      v3d.Transform.identity().as_xnp(tr.xnp),
  )
  identity_tr = identity_tr.broadcast_to(tr_shape)
  assert identity_tr.shape == tr_shape
  assert identity_tr.R.shape == tr_shape + (3, 3)
  assert identity_tr.t.shape == tr_shape + (3,)

  # Inverting the matrix is equivalent to the matrix of the invert transform
  dca.testing.assert_array_equal(tr.inv.matrix4x4, enp.compat.inv(tr.matrix4x4))

  # Inverting twice the transformation should be a no-op
  dca.testing.assert_array_equal(tr.inv.inv, tr)

  # Composing the transformation with the inverse should be identity
  dca.testing.assert_array_equal(tr.inv @ tr, identity_tr)
  dca.testing.assert_array_equal(tr @ tr.inv, identity_tr)
  dca.testing.assert_array_equal(tr @ identity_tr, tr)
  dca.testing.assert_array_equal(identity_tr @ tr, tr)

  # Exporting/importing matrix from 4x4 should be a no-op
  dca.testing.assert_array_equal(v3d.Transform.from_matrix(tr.matrix4x4), tr)

  dca.testing.assert_array_equal(tr + [0, 0, 0], tr)
  dca.testing.assert_array_equal(tr + tr.xnp.asarray([0, 0, 0]), tr)

  # Figure should work
  _ = tr.fig


def _assert_ray_transformed(
    tr: v3d.Transform,
    other_shape: dca.typing.Shape,
    expected_shape: dca.typing.Shape,
    test_values: TransformExpectedValue,
):
  """Test ray transformation."""
  xnp = tr.xnp
  ray = v3d.Ray(
      pos=xnp.asarray(_RAY_POS),
      dir=xnp.asarray(_RAY_DIR),
  )
  ray = ray.broadcast_to(other_shape)
  assert ray.shape == other_shape

  expected_ray = v3d.Ray(
      pos=xnp.asarray(test_values.expected_pos),
      dir=xnp.asarray(test_values.expected_dir),
  )
  expected_ray = expected_ray.broadcast_to(expected_shape)
  dca.testing.assert_array_equal(tr @ ray, expected_ray)


def _assert_point_transformed(
    tr: v3d.Transform,
    other_shape: dca.typing.Shape,
    expected_shape: dca.typing.Shape,
    test_values: TransformExpectedValue,
):
  """Test point transformation."""
  xnp = tr.xnp

  # Test transform point position
  expected_point_pos = xnp.asarray(test_values.expected_pos)
  expected_point_pos = xnp.broadcast_to(
      expected_point_pos,
      expected_shape + (3,),
  )

  point_pos = xnp.asarray(_RAY_POS)
  point_pos = xnp.broadcast_to(point_pos, other_shape + (3,))

  dca.testing.assert_array_equal(tr @ point_pos, expected_point_pos)
  dca.testing.assert_array_equal(tr.apply_to_pos(point_pos), expected_point_pos)

  # Test transform point direction
  expected_point_dir = xnp.asarray(test_values.expected_dir)
  expected_point_dir = xnp.broadcast_to(
      expected_point_dir,
      expected_shape + (3,),
  )

  point_dir = xnp.asarray(_RAY_DIR)
  point_dir = xnp.broadcast_to(point_dir, other_shape + (3,))

  dca.testing.assert_array_equal(tr.apply_to_dir(point_dir), expected_point_dir)


def _assert_tr_transformed(
    tr: v3d.Transform,
    other_shape: dca.typing.Shape,
    expected_shape: dca.typing.Shape,
    test_values: TransformExpectedValue,
):
  """Test transform transformation."""
  xnp = tr.xnp
  other_tr = v3d.Transform(
      R=xnp.asarray(_TR_R),
      t=xnp.asarray(_TR_T),
  )
  other_tr = other_tr.broadcast_to(other_shape)

  expected_tr = v3d.Transform(
      R=xnp.asarray(test_values.expected_r),
      t=xnp.asarray(test_values.expected_t),
  )
  expected_tr = expected_tr.broadcast_to(expected_shape)

  dca.testing.assert_array_equal(tr @ other_tr, expected_tr)

  # Composing transformations or matrix is equivalent
  # Only test for scalar shape as broadcasting & vectorization have different
  # rules
  if not tr.shape:
    dca.testing.assert_array_equal(
        v3d.Transform.from_matrix(tr.matrix4x4 @ other_tr.matrix4x4),
        expected_tr,
    )


def _assert_camera_transformed(
    tr: v3d.Transform,
    other_shape: dca.typing.Shape,
    expected_shape: dca.typing.Shape,
    test_values: TransformExpectedValue,
):
  """Test transform transformation."""
  xnp = tr.xnp
  spec = v3d.PinholeCamera.from_focal(
      resolution=(12, 12),
      focal_in_px=34,
  )
  spec = spec.as_xnp(xnp)
  cam = v3d.Camera(
      spec=spec,
      world_from_cam=v3d.Transform(
          R=xnp.asarray(_TR_R),
          t=xnp.asarray(_TR_T),
      ),
  )
  cam = cam.broadcast_to(other_shape)

  expected_cam = v3d.Camera(
      spec=spec,
      world_from_cam=v3d.Transform(
          R=xnp.asarray(test_values.expected_r),
          t=xnp.asarray(test_values.expected_t),
      ),
  )
  expected_cam = expected_cam.broadcast_to(expected_shape)

  dca.testing.assert_array_equal(tr @ cam, expected_cam)


def _assert_scale(
    tr: v3d.Transform,
    *,
    xnp: enp.NpModule,
    tr_shape: dca.typing.Shape,
    expected_r: FloatArray[3, 3],
    expected_scale_xyz: FloatArray[3],
    expected_scale: Optional[FloatArray['']],
):
  assert tr.xnp is xnp
  assert tr.shape == tr_shape
  expected_r = xnp.asarray(expected_r)
  expected_scale_xyz = xnp.asarray(expected_scale_xyz)
  np.testing.assert_allclose(
      tr.R,
      xnp.broadcast_to(expected_r, tr_shape + (3, 3)),
      atol=1e-6,
  )
  np.testing.assert_allclose(
      tr.scale_xyz,
      xnp.broadcast_to(expected_scale_xyz, tr_shape + (3,)),
  )
  if expected_scale is None:
    with pytest.raises(ValueError, match='Cannot get `Transform.scale`'):
      _ = tr.scale
  else:
    expected_scale = xnp.asarray(expected_scale)
    np.testing.assert_allclose(
        tr.scale,
        xnp.broadcast_to(expected_scale, tr_shape),
    )


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize(
    'tr_shape',
    [
        (),
        (5,),
    ],
)
def test_transformation_scale(
    xnp: enp.NpModule,
    tr_shape: dca.typing.Shape,
):
  tr = v3d.Transform.from_angle(x=xnp.asarray(enp.tau / 4))
  tr = tr.broadcast_to(tr_shape)

  if tr_shape and xnp in [
      enp.lazy.tnp,
      # Attempted to vmap over aten::_unique2
      enp.lazy.torch,
  ]:
    pytest.skip('Vectorization not supported yet with TF')

  assert_scale = functools.partial(_assert_scale, xnp=xnp, tr_shape=tr_shape)

  assert_scale(
      tr,
      expected_r=[
          [1, 0, 0],
          [0, 0, -1],
          [0, 1, 0],
      ],
      expected_scale_xyz=[1, 1, 1],
      expected_scale=1,
  )
  assert_scale(
      tr.mul_scale(3.0),
      expected_r=[
          [3, 0, 0],
          [0, 0, -3],
          [0, 3, 0],
      ],
      expected_scale_xyz=[3, 3, 3],
      expected_scale=3,
  )
  assert_scale(
      tr.mul_scale(3.0).normalize(),
      expected_r=[
          [1, 0, 0],
          [0, 0, -1],
          [0, 1, 0],
      ],
      expected_scale_xyz=[1, 1, 1],
      expected_scale=1,
  )
  assert_scale(
      tr.mul_scale([2, -3, 1.5]),
      expected_r=[
          [2, 0, 0],
          [0, 0, -1.5],
          [0, -3, 0],
      ],
      # TODO(epot): Detect scale sign
      expected_scale_xyz=[2, 3, 1.5],
      expected_scale=None,
  )


@enp.testing.parametrize_xnp()
def test_transformation_from_angle_multiple(xnp: enp.NpModule):
  # TODO(epot): Supports vectorization
  z = enp.tau / 4
  y = -enp.tau / 8
  x = xnp.asarray(enp.tau / 16)

  rz = v3d.math.rot_z(z)
  ry = v3d.math.rot_y(y)
  rx = v3d.math.rot_x(x)

  tr = v3d.Transform.from_angle(x=x, y=y, z=z)

  assert tr.xnp is xnp

  if xnp is enp.lazy.torch:  # Torch do not support `np @ Tensor`
    rz = xnp.asarray(rz, dtype=xnp.float32)
    ry = xnp.asarray(ry, dtype=xnp.float32)

  dca.testing.assert_array_equal(tr.R, rz @ ry @ rx)

  dca.testing.assert_array_equal(
      v3d.Transform.from_angle(),
      v3d.Transform.identity(),
  )


@enp.testing.parametrize_xnp()
def test_transformation_from_angle(xnp: enp.NpModule):
  # TODO(epot): Supports vectorization
  z = xnp.asarray(enp.tau / 4)
  y = xnp.asarray(-enp.tau / 8)
  x = xnp.asarray(enp.tau / 16)

  rz = v3d.math.rot_z(z)
  ry = v3d.math.rot_y(y)
  rx = v3d.math.rot_x(x)

  tr_x = v3d.Transform.from_angle(x=x)
  tr_y = v3d.Transform.from_angle(y=y)
  tr_z = v3d.Transform.from_angle(z=z)
  assert tr_x.xnp is xnp
  assert tr_y.xnp is xnp
  assert tr_z.xnp is xnp

  dca.testing.assert_array_equal(tr_x.R, rx)
  dca.testing.assert_array_equal(tr_y.R, ry)
  dca.testing.assert_array_equal(tr_z.R, rz)

  # Identity
  dca.testing.assert_array_equal(
      v3d.Transform.from_angle(),
      v3d.Transform.identity(),
  )



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests."""

from etils import enp
import numpy as np
import pytest
import visu3d as v3d

# Activate the fixture
enable_tf_np_mode = enp.testing.set_tnp


@enp.testing.parametrize_xnp()
@pytest.mark.parametrize(
    'value, shape',
    [
        (
            [
                [2.0, 4, 0.3],
                [-2.0, -4, -0.3],
            ],
            (2,),
        ),
        (
            [2.0, 4, 0.3],
            (),
        ),
    ],
)
def test_spherical_coordinates(xnp: enp.NpModule, value, shape):
  pts = xnp.asarray(value)

  spherical = v3d.math.carthesian_to_spherical(pts)

  assert enp.lazy.get_xnp(spherical.r) is xnp
  assert enp.lazy.get_xnp(spherical.theta) is xnp
  assert enp.lazy.get_xnp(spherical.phi) is xnp
  assert spherical.r.shape == shape
  assert spherical.theta.shape == shape
  assert spherical.phi.shape == shape

  pts_round_trip = v3d.math.spherical_to_carthesian(*spherical)
  assert enp.lazy.get_xnp(pts_round_trip) is xnp
  assert pts_round_trip.shape == shape + (3,)

  np.testing.assert_allclose(pts, pts_round_trip, atol=1e-6)



# Copyright 2022 The visu3d Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Point3d utils."""

from __future__ import annotations

import dataclasses
import typing
from typing import Optional, Union

import dataclass_array as dca
from etils.array_types import FloatArray, ui8  # pylint: disable=g-multiple-import
from visu3d import array_dataclass
from visu3d import plotly
from visu3d.dc_arrays import transformation
from visu3d.plotly import fig_config_utils
from visu3d.utils import np_utils
from visu3d.utils.lazy_imports import plotly_base

if typing.TYPE_CHECKING:
  from visu3d.dc_arrays import camera_spec as camera_spec_lib


class Point3d(array_dataclass.DataclassArray):
  """3d point cloud.

  Attributes:
    p: 3d (x, y, z) coordinates
    rgb: uint8 color
  """

  p: FloatArray['*shape 3']
  rgb: Optional[ui8['*shape 3']] = None

  # Overwrite `v3d.DataclassArray.fig_config`.
  fig_config: plotly.TraceConfig = dataclasses.field(
      default=plotly.TraceConfig(
          num_samples=fig_config_utils.LazyValue(
              lambda fig_config: fig_config.num_samples_point3d
          )
      ),
      repr=False,
      init=False,
  )

  def __add__(self, translation: FloatArray['... 3']) -> Point3d:
    """Translate the position."""
    translation = self.xnp.asarray(translation)
    return self.replace(p=self.p + translation)

  __sub__ = np_utils.__sub__

  def clip(
      self,
      min: Optional[Union[FloatArray['3'], float]] = None,  # pylint: disable=redefined-builtin
      max: Optional[Union[FloatArray['3'], float]] = None,  # pylint: disable=redefined-builtin
  ) -> Point3d:
    """Clip the position coordinates to the (min, max) boundaries."""
    min = dca.utils.np_utils.asarray(min, xnp=self.xnp, optional=True)
    max = dca.utils.np_utils.asarray(max, xnp=self.xnp, optional=True)
    return self.replace(p=self.p.clip(min, max))

  # Protocols (inherited)

  def apply_transform(self, tr: transformation.Transform) -> Point3d:
    # No `color` modification
    return self.replace(p=tr.apply_to_pos(self.p))

  def apply_px_from_cam(self, spec: camera_spec_lib.CameraSpec) -> Point2d:
    """Apply the `px_from_cam @ self` transformation."""
    # TODO(epot): Expose with_depth in the public API
    p, depth = spec._px_and_depth_from_cam(self.p)  # pylint: disable=protected-access
    return Point2d(
        p=p,
        rgb=self.rgb,
        depth=depth,
    )

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    return plotly.make_points(self.p, color=self.rgb)


class Point2d(array_dataclass.DataclassArray):
  """2d point cloud.

  Attributes:
    p: 2d (x, y) == (u, v) == (w, h) coordinates
    rgb: uint8 color
    depth: The depth in camera coordinates.
  """

  p: FloatArray['*shape 2']
  depth: Optional[FloatArray['*shape 1']] = None
  rgb: Optional[ui8['*shape 3']] = None

  # Overwrite `v3d.DataclassArray.fig_config`.
  fig_config: plotly.TraceConfig = dataclasses.field(
      default=plotly.TraceConfig(
          num_samples=fig_config_utils.LazyValue(
              lambda fig_config: fig_config.num_samples_point2d
          )
      ),
      repr=False,
      init=False,
  )

  def clip(
      self,
      min: Optional[Union[FloatArray['2'], float]] = None,  # pylint: disable=redefined-builtin
      max: Optional[Union[FloatArray['2'], float]] = None,  # pylint: disable=redefined-builtin
  ) -> Point3d:
    """Clip the position coordinates to the (min, max) boundaries."""
    min = dca.utils.np_utils.asarray(min, xnp=self.xnp, optional=True)
    max = dca.utils.np_utils.asarray(max, xnp=self.xnp, optional=True)
    return self.replace(p=self.p.clip(min, max))

  # Protocols

  def apply_cam_from_px(
      self,
      spec: camera_spec_lib.CameraSpec,
  ) -> Point3d:
    """Apply the `px_from_cam @ self` transformation."""
    p = spec.cam_from_px @ self.p
    if self.depth is not None:
      p = p * self.depth
    return Point3d(
        p=p,
        rgb=self.rgb,
    )

  def make_traces(self) -> list[plotly_base.BaseTraceType]:
    traces = []

    (trace,) = plotly.make_points(self.p, color=self.rgb)
    traces.append(trace)

    if self.depth is not None:
      pass  # TODO(epot): Add depth trace (no displayed by default)
      # trace, = plotly.make_points(
      #     self.p,
      #     color=self.depth
      # )
      # traces.append(trace)

    return traces
